from level_io import level_path, instantiate, Params
from level import Level

class LevelError(EnvironmentError):
	pass


class Reader(object):
	#  line     current line
	#  lineno   current line number (1-based)

	def __init__(self, f, filename):
		self.f = f
		self.filename = filename
		self.lineno = 0
		self.next_line()
	
	def next_line(self):
		self.lineno += 1
		line = self.f.readline()
		if not line:
			self.error("Unexpected end of file")
		self.line = line.strip()
	
	def splitline(self, sep = None):
		self.tokens = self.line.split(sep)
		return self.get(str)
	
	def get(self, *types):
		if len(self.tokens) < len(types):
			self.error("Not enough tokens")
		result = []
		for t in types:
			s = self.tokens.pop(0)
			try:
				v = t(s)
			except ValueError:
				self.error("Type error")
			result.append(v)
		if len(result) == 1:
			return result[0]
		else:
			return result
	
	def expect(self, what):
		kwd = self.splitline()
		if kwd <> what:
			self.error("Expected '%s'" % what)		
	
	def part(self, s, sep):
		parts = s.split(sep, 1)
		#print "Reader.part: parts =", parts
		if len(parts) <> 2:
			self.error("Missing '%s'" % sep)
		return parts
	
	def eval(self, expr):
		try:
			return eval(expr)
		except Exception, e:
			self.error("%s: %s" % (e.__class__.__name__, e))
	
	def error(self, mess):
		raise LevelError("%s, line %s: %s" % (self.filename, self.lineno, mess))


def read_array(r, a):
	nmap = r.get(int)
	r.next_line()
	mapping = {}
	mapping["0"] = None
	while nmap:
		index, desc = r.part(r.line, ":")
		cls, tail = r.part(desc, "(")
		if not tail.endswith(")"):
			r.error("Missing ')'")
		parms = tail[:-1]
		#def splitparm(parm):
		#	s = r.part(parm, "=")
		#	if len(s) <> 2:
		#		r.error("Malformed parameter")
		#	return s
		#p = Params()
		#for name, value in map(splitparm, parms.split(",")):
		#	setattr(p, name.strip(), r.eval(value))
		p = r.eval("Params(%s)" % parms)
		x = instantiate(cls.strip(), p)
		if not x:
			r.error("Unknown class '%s'" % cls)
		mapping[index] = x
		r.next_line()
		nmap -= 1
	nrows, ncols = a.shape
	for row in xrange(nrows):
		indices = r.line.split()
		if len(indices) <> ncols:
			r.error("Wrong number of items in row")
		for col in xrange(ncols):
			x = mapping.get(indices[col])
			if x:
				a[row, col] = x.clone()
		r.next_line()

def load_level(filename):
	path = level_path(filename)
	f = open(path, "r")
	r = Reader(f, path)
	r.expect("Level")
	nrows, ncols = r.get(int, int)
	level = Level(nrows, ncols)
	r.next_line()
	while r.line <> "EndLevel":
		kwd = r.splitline()
		if kwd == "BackLayer":
			read_array(r, level.backs)
		elif kwd == "WallLayer":
			read_array(r, level.walls)
		elif kwd == "FloorLayer":
			read_array(r, level.floors)
		elif kwd == "GizmoLayer":
			read_array(r, level.gizmos)
		elif kwd == "ObjectLayer":
			read_array(r, level.objects)
	r.expect("EndLevel")
	level.update_tags()
	return level
